#ifndef gltex_atlas_vert_H
#define gltex_atlas_vert_H
#include <stddef.h>
extern const char *const _binary_gltex_atlas_vert_start;
extern const char *const _binary_gltex_atlas_vert_end;
extern size_t _binary_gltex_atlas_vert_size;
#endif // gltex_atlas_vert_H